package com.maroontress.coverture;

import com.maroontress.coverture.gcda.DataRecord;
import com.maroontress.coverture.gcda.FunctionDataRecord;
import com.maroontress.coverture.gcno.FunctionGraphRecord;
import com.maroontress.coverture.gcno.NoteRecord;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.Collection;
import java.util.Comparator;
import java.util.TreeMap;

/**
   gcnoեѡ̤ݻޤ
*/
public final class Note {

    /** gcnoեΥꥸǤ */
    private Origin origin;

    /** gcnoեΥСֹǤ */
    private int version;

    /**
       gcnoեΥॹפǤgcdaեƱȤƤ
       뤳Ȥǧ뤿˻Ѥޤ
    */
    private int stamp;

    /** ؿդȤμ̻ҤΥޥåפǤ */
    private TreeMap<Integer, FunctionGraph> map;

    /** Ϣ륽ɤΥꥹȤǤ */
    private SourceList sourceList;

    /** ץμ¹ԲǤ */
    private int runs;

    /** ץοǤ */
    private int programs;

    /**
       Ρȥ쥳ɤ饤󥹥󥹤ޤ

       @param rec Ρȥ쥳
       @param origin gcnoեΥꥸ
       @throws CorruptedFileException եι¤Ƥ뤳Ȥ򸡽
    */
    private Note(final NoteRecord rec, final Origin origin)
	throws CorruptedFileException {
	this.origin = origin;
	version = rec.getVersion();
	stamp = rec.getStamp();
	map = new TreeMap<Integer, FunctionGraph>();
	sourceList = new SourceList();

	FunctionGraphRecord[] list = rec.getList();
	for (FunctionGraphRecord e : list) {
	    FunctionGraph fg = new FunctionGraph(e);
	    map.put(fg.getId(), fg);
	}
    }

    /**
       gcovߴΥեΥХåޤ

       @param prop ϥץѥƥ
    */
    public void createSourceList(final IOProperties prop) {
	sourceList.outputFiles(origin, runs, programs, prop);
    }

    /**
       եΥꥹȤ򹹿ޤgcdaեѡ
       ˸ƤӽФɬפޤ
    */
    private void updateSourceList() {
	Collection<FunctionGraph> all = map.values();
	for (FunctionGraph g : all) {
	    g.addLineCounts(sourceList);
	}
    }

    /**
       ΡȤXMLǽϤޤ

       @param out 
    */
    public void printXML(final PrintWriter out) {
	File file = origin.getNoteFile();
	out.printf("<note file='%s' version='0x%x' stamp='0x%x'"
		   + " lastModified='%d'>\n",
		   XML.escape(file.getPath()), version, stamp,
		   file.lastModified());
	sourceList.printXML(out);
	Collection<FunctionGraph> allGraph = map.values();
	for (FunctionGraph g : allGraph) {
	    g.printXML(out);
	}
	out.printf("</note>\n");
    }

    /**
       ǡ쥳ɤꤷޤ

       @param rec ǡ쥳
       @param file gcdaե
       @throws CorruptedFileException եι¤Ƥ뤳Ȥ򸡽
    */
    private void setDataRecord(final DataRecord rec, final File file)
	throws CorruptedFileException {
	String path = origin.getNoteFile().getPath();
	if (version != rec.getVersion()) {
	    throw new CorruptedFileException(path + ": version mismatch.");
	}
	if (stamp != rec.getStamp()) {
	    throw new CorruptedFileException(path + ": timestamp mismatch.");
	}
	if (origin.getNoteFile().lastModified() > file.lastModified()) {
	    System.err.printf(
		"%s: warning: gcno file is newer than gcda file.%n", path);
	}
	FunctionDataRecord[] list = rec.getList();
	for (FunctionDataRecord e : list) {
	    int id = e.getId();
	    FunctionGraph g = map.get(id);
	    if (g == null) {
		System.err.printf(
		    "%s: warning: unknown function id '%d'.%n", path, id);
		continue;
	    }
	    g.setFunctionDataRecord(e);
	}
	runs = rec.getObjectSummary().getRuns();
	programs = rec.getProgramSummaries().length;
    }

    /**
       gcdaեѡơΡȤ˥󥿤ɲäޤ
       ͥޥåפΤǡ2GХȤĶեϰޤ

       եƤʾϡɸ२顼Ϥ˥åȥ졼
       Ϥޤ

       @param origin gcnoեΥꥸ
    */
    private void parseData(final Origin origin) {
	File dataFile = origin.getDataFile();
	RandomAccessFile file;
	try {
	    file = new RandomAccessFile(dataFile, "r");
	} catch (FileNotFoundException e) {
	    System.err.printf("%s: not found.%n", dataFile.getPath());
	    return;
	}
	FileChannel ch = file.getChannel();
	try {
	    try {
		ByteBuffer bb = ch.map(FileChannel.MapMode.READ_ONLY,
				       0, ch.size());
		DataRecord dataRecord = new DataRecord(bb);
		setDataRecord(dataRecord, dataFile);
	    } catch (CorruptedFileException e) {
		e.printStackTrace();
	    } finally {
		file.close();
	    }
	} catch (IOException e) {
	    e.printStackTrace();
	}
    }

    /**
       gcnoեѡơΡȤޤͥޥå
       Τǡ2GХȤĶեϰޤ

       եƤʾϡɸ२顼Ϥ˥åȥ졼
       Ϥơnull֤ޤ

       @param path gcnoեΥѥ
       @return Ρ
       @throws IOException ϥ顼
    */
    public static Note parse(final String path) throws IOException {
	if (!path.endsWith(".gcno")) {
	    System.err.printf("%s: suffix is not '.gcno'.%n", path);
	    return null;
	}
	Origin origin = new Origin(path);
	RandomAccessFile file;
	try {
	    file = new RandomAccessFile(path, "r");
	} catch (FileNotFoundException e) {
	    System.err.printf("%s: not found.%n", path);
	    return null;
	}
	FileChannel ch = file.getChannel();
	Note note = null;
	try {
	    try {
		ByteBuffer bb = ch.map(FileChannel.MapMode.READ_ONLY,
				       0, ch.size());
		NoteRecord noteRecord = new NoteRecord(bb);
		note = new Note(noteRecord, origin);
	    } catch (CorruptedFileException e) {
		e.printStackTrace();
		return null;
	    } finally {
		file.close();
	    }
	} catch (IOException e) {
	    e.printStackTrace();
	    return null;
	}
	note.parseData(origin);
	note.updateSourceList();
	return note;
    }

    /**
       ꥸӤ륳ѥ졼Ǥ
    */
    private static Comparator<Note> originComparator;

    static {
	originComparator = new Comparator<Note>() {
	    public int compare(final Note n1,
			       final Note n2) {
		return n1.origin.compareTo(n2.origin);
	    }
	};
    }

    /**
       ꥸӤ륳ѥ졼֤ޤ

       @return ꥸӤ륳ѥ졼
    */
    public static Comparator<Note> getOriginComparator() {
	// ComparableƤ⤤...
	return originComparator;
    }
}
